﻿zwave_commands = {
	add_node = 1,
	remove_node = 2,
	remove_failed_node = 3,
	refresh_node_info = 4,
	request_node_state  = 5,
	request_node_dynamic = 6,
	set_node_manufacturer_name = 7,
	set_node_product_name  = 8,
	set_node_name = 9,
	set_node_location = 10,
	set_value_label = 11,
	set_value_units = 12,
	set_value_help = 13,
	set_value = 14,
	refresh_value = 15,
	set_change_verified = 16,
    reset_controller = 17,
    soft_reset = 18,
    cancel_controller_command = 19,
    test_network_node  = 20,
    test_network  = 21,
    heal_network_node  = 22,
    heal_network  = 23,
    request_node_neighbor_update  = 24,
    assign_return_route  = 25,
    delete_all_return_routes  = 26, 
    write_config  = 27,
    set_poll_interval = 28,
    enable_poll = 29,
    disable_poll = 30,
    set_poll_intensity = 31,
    press_button = 32,
    release_button = 33,
    refresh_mapping = 50
}

redis = require 'redis'
rclient = redis.connect()

_rname_driver = "zwave:driver"
_rname_notifications = "zwave:notifications"
function _rname_present(home_id) return "zwave:present:"..home_id end 
function _rname_node(home_id) return "zwave:"..home_id..":node" end 
function _rname_value(home_id, node_id, value_id) return "zwave:"..home_id..":value:"..node_id..":"..value_id end 
function _rname_node_values(home_id, node_id) return "zwave:"..home_id..":node:"..node_id..":values" end 
function _rname_address(address) return "zwave:address:"..address end 
function _rname_address_hkey(home_id, node_id, value_id, instance) return home_id.."_"..node_id.."_"..value_id.."_"..instance end 

function _set_row(name, key, fields)
	return rclient:hset(name, key, json.encode(fields))
end
function _get_row(name, key)
	local d = rclient:hget(name, key)
	if d then d = json.pdecode(d) end 
	return d 
end
function _update_row(name, key, fields, create)
	local d = _get_row(name, key)
	if d then 
		for k, v in pairs(fields) do d[k]=v end 
	elseif create then 
		d = fields 
	end 
	if d then _set_row(name, key, d) end 
	return d 
end

function _get_driver(home_id)
	local result = _get_row(_rname_driver, home_id)
	if result then result.home_id=home_id end 
	return result 
end 

function _get_node(node_id, home_id) 
	local result=_get_row(_rname_node(home_id), node_id) 
	if result then result.node_id=node_id end 
	return result 
end 

function _get_value(home_id, node_id, value_id, instance) 
	local result=_get_row(_rname_value(home_id, node_id, value_id), instance)
	if result then 
		result.node_id=node_id 
		result.value_id=value_id 
	end  
	return result 
end 

function _remove_address(home_id, node_id, id, instance, address) 
	rclient:hdel(
		_rname_address(address), 
		_rname_address_hkey(home_id, node_id, id, instance)
	)
end 

function _add_address(home_id, node_id, id, instance, address, is_readonly) 
	rclient:hset(
		_rname_address(address), 
		_rname_address_hkey(home_id, node_id, id, instance),
		json.encode({id=id, home_id=home_id, node_id=node_id, is_readonly=is_readonly, instance=instance})
	)
end 

function _update_value(home_id, node_id, value_id, instance, params) 
	_update_row(_rname_value(home_id, node_id, value_id), instance, params)
end

function _get_values(node_id, home_id)
	result = {} 
	for _, value_id in ipairs(rclient:smembers(_rname_node_values(home_id, node_id))) do 
		for instance, value in pairs(rclient:hgetall(_rname_value(home_id, node_id, value_id))) do 
			value = json.pdecode(value)
			if type(value)=="table" then 
				value.id = value_id
				value.instance =instance
				table.insert(result, value)
			end 
		end 
	end 
	return result 
end 

function get_home_id(onlyDB) 
	local result = 0 
	local home_id 
	if not onlyDB then 
		home_id = getvar("hid") 
		if home_id then home_id=tonumber(home_id) end 
	end 
	if type(home_id)~="number" then 
		local keys = rclient:hkeys(_rname_driver)	
		for _, k in ipairs(keys) do 
			k=tonumber(k)
			if k~=0 then result= k end 	
		end 
	else 
		if rclient:hexists(_rname_driver, home_id) then result = home_id end 
	end 
	return result
end 

function get_notification() 
	return rclient:rpop(_rname_notifications)
end 

function is_active_daemon(home_id) 
	return rclient:exists(_rname_present(home_id))
end 

function get_driver(home_id) 
	local driver = _get_driver(home_id)
	if driver then 
		driver.active_daemon = is_active_daemon(home_id) 
	else 
		driver = false 
	end 
	return driver 
end 

function get_nodes(home_id)
	if not home_id then home_id=_get_home_id()  end 
	result = {} 
	if home_id then
		local t = rclient:hgetall(_rname_node(home_id))
		for k, r in pairs(t) do 
			r = json.pdecode(r)
			if r then 
				r.node_id = k 
				table.insert(result, r)
			end 
		end 
	end  
	return result 
end 

function get_node(node_id, home_id)
	local node = _get_node(node_id, home_id)
	if node then 
		node.values = {user={}, config={}} 
		local values = _get_values(node_id, home_id)
		if values then 
			for _, v in ipairs(values) do
				local genre = "user"
				if v.genre == "config" or 
					v.genre == "system" then 
					genre = "config" 
				end 
			--	if v.datatype == "list" then genre = "user" end 

				if v.genre then 
					if v.address and type(v.address)=="number" and v.address>0 then 
						v.address = knxlib.decodega(v.address) 
						local o = grp.find(v.address) 
						if o then 
							v.address_text = o.name and o.name .. " (" .. o.address .. ")" or o.address
						else 
							v.address = "" 
						end 
					else 
						v.address = "" 
					end 
					table.insert(node.values[genre], v)
				end 
			end 
		end 
	end 
	return node	
end 

function set_basic_node(node_id, home_id, params) 
	local node = _get_node(node_id, home_id) 
	if node then 
		local values = params.objects or {} 	
		--Values и их Объекты 
		for i, v in pairs(params.objects) do 
			local id = tonumber(v.id)
			local instance = tonumber(v.instance)

			local value
			if id and instance then value = _get_value(home_id, node_id, id, instance) end 

			if value then 
				local is_writeonly = value.is_writeonly and value.is_writeonly==1 and true or false 
				local is_writebus = value.is_writebus and value.is_writebus==1 and true or false 
				local set_is_writebus 
				if not is_writeonly then 
					set_is_writebus = v.is_writebus and 1 or 0  
				end 
				local address 
				--создаем адрес 
				if v.address == "new" then 
					local name = #params.name>0 and params.name
						or #node.name>0 and node.name 
						or node.product_name or node.manufacturer_name or "" 
					local object = 	{
						name = name .. " - " .. value.label, 
						tags = {'zwave'}, 
						units = value.units or ""
					}
					local datatype = zwave_dt_to_knx(value.datatype) or dt.string
					object.datatype = datatype 
						
					address = grp.create(object) 
					if address then 
						local a = knxlib.encodega(address) 
						if a then
							value.address = tonumber(value.address)
							if value.address and value.address>0 then 
								_remove_address(home_id, node_id, id, instance, value.address) 
							end 
							if value.datatype=="list" then _list_object(value, a) end 
							_update_value(home_id, node_id, id, instance, {address=a, is_writebus=set_is_writebus}) 
							_add_address(home_id, node_id, id, instance, a, value.is_readonly) 
						else 
							address = nil 
						end 
					end 
				elseif v.address == "remove"  then 
					value.address = tonumber(value.address)
					if value.address and value.address>0 then
						_remove_address(home_id, node_id, id, instance, value.address) 
						_update_value(home_id, node_id, id, instance, {address=0, is_writebus=0})
					end 
				else 
					address = v.address
					if type(address) == "string" and #address>0 then 
						local a = knxlib.encodega(address) 
						if a and value.address~=a then 
							if value.datatype=="list" then _list_object(value, a) end 
							if value.address and value.address>0 then
								_remove_address(home_id, node_id, id, instance, value.address) 
							end 
							_update_value(home_id, node_id, id, instance, {address=a, is_writebus=set_is_writebus})
							_add_address(home_id, node_id, id, instance, a, value.is_readonly) 
						elseif set_is_writebus~=nil then 
							address = nil 
							_update_value(home_id, node_id, id, instance, {is_writebus=set_is_writebus})
						else 
							address = nil 
						end 
					else 
						address = nil 
					end 
						end 

				--опрос переменной 
				if not is_writeonly then 
					value.is_value_polled = value.is_value_polled and value.is_value_polled==1 and true or false 
					v.is_value_polled = toboolean(v.is_value_polled)
					if v.is_value_polled ~= nil and v.is_value_polled ~= value.is_value_polled then 
						_zwave_command({
							cmd = v.is_value_polled and zwave_commands.enable_poll or zwave_commands.disable_poll, 
							node_id = node_id, 
							value_id = id,
							instance = instance
						})
						value.is_value_polled = v.is_value_polled
					end 
					if value.is_value_polled then 
						local polled_intensity = value.polled_intensity
						if v.polled_intensity~=nil then 
							v.polled_intensity = tonumber(v.polled_intensity)
							if value.polled_intensity~=v.polled_intensity then 
								_zwave_command({
									cmd = zwave_commands.set_poll_intensity, 
									node_id = node_id, 
									value_id = id,
									instance = instance, 
									intensity = v.polled_intensity
								})
							end 
						end 
					end 
				end 
				
				--пишем текущее значение 
				if address and 	set_is_writebus~=nil then 
					local v = json.pdecode(value.value)
					if type(v)=="table" then v = tonumber(_zwave_list_value(v)) end  --для list 
					if v~=nil then 
						if set_is_writebus == 1 then 
							grp.write(address, v)
						else 
							grp.update(address, v)
						end 
					end 
				end 
			end 
		end 

		--имя 
		if type(params.name) == "string" and #params.name<100 and params.name~=node.name then 
			_zwave_command({
				cmd = zwave_commands.set_node_name, 
				node_id = node_id, 
				name = params.name,
				instance = instance 
			})
		end 		
		-- расположени 
		if type(params.location) == "string" and #params.location<100 and params.location~=node.location then 
			_zwave_command({
				cmd = zwave_commands.set_node_location, 
				node_id = node_id, 
				location = params.location,
				instance = instance
			})
		end 		
						
		node = get_node(node_id, home_id) 
	end 
	
	return node 
end 

function add_node() 
	_zwave_command({
		cmd = zwave_commands.add_node, 
		do_security = true,
	})
end 

function cancel_command() 
	_zwave_command({
		cmd = zwave_commands.cancel_controller_command
	})
end 

function remove_node() 
	_zwave_command({
		cmd = zwave_commands.remove_node
	})
end 

function remove_node_failed(node_id) 
	_zwave_command({
		cmd = zwave_commands.remove_failed_node, 
		node_id = node_id,
	})
	return true 
end 

function  set_extended_node(node_id, home_id, params) 
	local node = _get_node(node_id, home_id) 
	if node then 
		local values = params.objects or {} 
		for i, v in pairs(params.objects) do 
			local id = tonumber(v.id)
			local instance = tonumber(v.instance)
			local value 
			if id then value = _get_value(home_id, node_id, id, instance) end 
			if value and value.datatype~="button" then
				value.value = _get_value_zwave(value)
				v.value = _get_value_request(value, v.value)
				if value.value~=nil and v.value~=nil and not _compare_value(value.value, v.value, value.datatype) then 
					_zwave_command({
						cmd = zwave_commands.set_value, 
						node_id = node_id, 
						value_id = id, 
						instance = instance,
						value = value.datatype=="list" and v.value.value or v.value
					})
				end 
			end 
		end 
		node = get_node(node_id, home_id) 
	end 
	return node
end 
	
function press_button(home_id, node_id, value_id, instance) 
	_zwave_command({
		cmd = zwave_commands.press_button,
		node_id = node_id,
		value_id = value_id, 
		instance = instance
	})
	return 1 
end 

function release_button(home_id, node_id, value_id, instance) 
	_zwave_command({
		cmd = zwave_commands.release_button,
		node_id = node_id,
		value_id = value_id, 
		instance = instance
	})
	return 1 
end 


function _zwave_command(cmd) 
	rclient:publish("zwave:commands", json.encode(cmd))
end 	
			
function _get_value_zwave(value) 
	local v  = json.pdecode(value.value) 
	local res 
	local datatype = value.datatype
	if v then
		if datatype=="byte" or datatype=="int" or datatype=="short" then 
			v = tonumber(v) 
		elseif  datatype=="decimal" then 
			v = tonumber(v) 
		elseif  datatype=="string" then 
			v = tostring(v) 
		elseif datatype=="bool" then 
			if type(v)~="boolean" then 
				v = tonumber(v) 
				v = v and v==1 and true or false 		
			end 
		elseif datatype=="list" then 
			local va = {} 
			for i, j in pairs(v) do 
				va = {id=tonumber(i), value=j} 
			end 
			v = va 
		end 		
	end 
	return v  
end 

function _get_value_request(value, v) 
	local datatype = value.datatype
	if type(v)~=nil then
		if datatype=="byte" or datatype=="int" or datatype=="short" then 
			v = tonumber(v) 
			local mi, ma = value.min, value.max 
			if v and math.ceil(v)==v and v>=mi and v<=ma then 
				res = v 
			end 		
		elseif  datatype=="decimal" then 
			local mi, ma = value.min, value.max 
			v = tonumber(v) 
			if v and v>=mi and v<=ma then 
				res = v 
			end 
		elseif  datatype=="string" then 
			res = tostring(v) 
		elseif datatype=="bool" then 
			if type(v)~="boolean" then 
				if v==1 then 
					res = true 
				elseif v==0 then
					res = false 
				end 
			else 
				res = v 
			end 
		elseif datatype=="list" then 
			local items = json.pdecode(value.items) 
			if type(items)=="table" then 
				for i, j in pairs(items) do 
					if j==v then 
						res = {id=i, value=j}
						break 
					end 
				end 
			end 
		end 		
	end 
	return res
end 

function _compare_value(v1, v2, datatype) 
	if datatype == "list" then 
		return tonumber(v1.id)==tonumber(v2.id) and v1.value==v2.value
	end 
	return v1 == v2 
end 

function zwave_dt_to_knx(datatype) 
	local res 
	if datatype=="bool" or  datatype=="button"  then 
		res = dt.bool 
	elseif datatype=="decimal" then 
		res = dt.float32		
	elseif datatype=="byte" then 
		res = dt.uint8
	elseif datatype=="short" then 
		res = dt.int16 
	elseif datatype=="int" then 
		res = dt.int32
	elseif datatype=="list" then 
		res = dt.int32 
	elseif datatype then 
		res = dt.string 
	end 
	return res 
end 

function zwave_stop_daemon() 
	os.execute("killall zwave")
	os.sleep(3)
	os.execute("killall -9 zwave")
	os.sleep(2)
	return 1
end

function zwave_change_frequency(f)
	if f =="eu" then 
		io.popen([[echo -ne "\x01\x04\x00\xF2\x00\x09" > /dev/ttyACM0]])
	elseif f=="ru" then 
		io.popen([[echo -ne "\x01\x04\x00\xF2\x01\x08" > /dev/ttyACM0]])
	elseif f == "us" then 
		io.popen([[echo -ne "\x01\x04\x00\xF2\x03\x06" > /dev/ttyACM0]])
	elseif f == "in" then 
		io.popen([[echo -ne "\x01\x04\x00\xF2\x02\x07" > /dev/ttyACM0]])
	end
	os.sleep(2) 	
end 

function zwave_clean_db()	
	for _, k in ipairs(rclient:keys("zwave:*")) do 
		rclient:del(k)
	end
	home_id=0
	rclient:rpush(_rname_notifications, json.encode({a="driver"}))
end 

function zwave_start_daemon()
	os.execute("killall zwave")
	os.sleep(3)
	os.execute("killall -9 zwave")
	os.sleep(2)
	os.execute("/home/apps/usr/sbin/zwave /home/apps/usr/share/openzwave/ /tmp/log/ /dev/ttyACM0 /home/apps/usr/share/zwave/handler.lua")
	rclient:rpush(_rname_notifications, json.encode({a="driver"}))
	os.sleep(5)
	return 1
end 

--добавляем vis параметры для объектов enum (list)
function _list_object(value, address) 
	local params = {
		visparams = [[{"control":"vselect","icons":"1"}]],
		enums = {}
	}
	local items = value.items 
	if type(items)=="string" then items = json.pdecode(items) end 
	if type(items)~="table" then  items = {}  end 
	for k, v in pairs(items) do 
		params.enums[tostring(k)]=v 
	end 	
	
	local val = value.value 
	if type(val)=="string" then val = json.pdecode(val) end 
	if type(val)=="table" then 
		params.enums.default = "Bad value"
	end 

	params.enums = json.encode(params.enums)

	db:update("objects", params, {address=address})
end 

function _zwave_list_value(value) 
	local result 
	for k, v in pairs(value) do
		result = tostring(k)
		break 
	end 
	return result 
end 