require('dbenv')
require('json')
require("apps")
require('genohm-scada.eibdgm')
db = dbenv:new()
redis = require 'redis'
rclient = redis.connect()
hid = nil

--[[
zwave:driver  - хеш с настройками драйвера
zwave:homeID:node -хеш со списков нод
zwave:homeID:value:node_id:value_id -хеш со списков instance
zwave:homeID:node:node_id:values -множество с ID value - денормализация для поиска
zwave:command -массив для обмена командами между демоном и интерфейсом
zwave:address:_address_ -hash c маппингом {home_id, node_id, value_id, instance, read_only}
]]

_rname_driver = "zwave:driver"
_rname_notifications = "zwave:notifications"
function _rname_present(home_id) return "zwave:present:"..home_id end
function _rname_node(home_id) return "zwave:"..home_id..":node" end
function _rname_value(home_id, node_id, value_id) return "zwave:"..home_id..":value:"..node_id..":"..value_id end
function _rname_node_values(home_id, node_id) return "zwave:"..home_id..":node:"..node_id..":values" end
function _rname_address(address) return "zwave:address:"..address end
function _rname_address_hkey(home_id, node_id, value_id, instance) return home_id.."_"..node_id.."_"..value_id.."_"..instance end

function _get_time() return os.time() end

function _notifications(data)
	rclient:rpush(_rname_notifications, json.encode(data))
	rclient:ltrim(_rname_notifications, -10, -1)
end

--перенос данных из старой версии в новую
function create_table_zwave()
	local isExist = db:getrow("SELECT * FROM sqlite_master WHERE type='table' AND name='zwave_driver'")
	if isExist then
		for _, row in pairs(db:getall("SELECT * FROM zwave_driver")) do
			if not rclient:hexists(_rname_driver, row.home_id) then
				rclient:hset(_rname_driver, row.home_id, json.encode(row))
			end
		end
		db:query("DROP TABLE zwave_driver")
	end
end
function create_table_nodes()
	local isExist = db:getrow("SELECT * FROM sqlite_master WHERE type='table' AND name='zwave_nodes'")
	if isExist then
		for _, row in pairs(db:getall("SELECT * FROM zwave_nodes")) do
			local name = _rname_node(row.home_id)
			if not rclient:hexists(name, row.node_id) then
				rclient:hset(name, row.node_id, json.encode(row))
			end
		end
		db:query("DROP TABLE zwave_nodes")
	end
end
function create_table_values()
	local isExist = db:getrow("SELECT * FROM sqlite_master WHERE type='table' AND name='zwave_values'")
	if isExist then
		for _, row in pairs(db:getall("SELECT * FROM zwave_values")) do
			local name = _rname_value(row.home_id, row.node_id, row.id)
			if not row.instance then  row.instance = 1 end
			if not rclient:hexists(name, row.instance) then
				rclient:hset(name, row.instance, json.encode(row))
			end
			rclient:sadd(_rname_node_values(row.home_id, row.node_id), row.id)
			if row.address then
				rclient:hset(
					_rname_address(row.address),
					_rname_address_hkey(row.home_id, row.node_id, row.id, row.instance),
					json.encode({id=row.id, home_id=row.home_id, node_id=row.node_id, is_readonly=row.is_readonly, instance=row.instance})
				)
			end
		end
		db:query("DROP TABLE zwave_values")
	end
end
function create_table_commands()
	local isExist = db:getrow("SELECT * FROM sqlite_master WHERE type='table' AND name='zwave_commands'")
	if isExist then
		db:query("DROP TABLE zwave_commands")
	end
end

commands = {
	add_node = 1,
	remove_node = 2,
	remove_failed_node = 3,
	refresh_node_info = 4,
	request_node_state  = 5,
	request_node_dynamic = 6,
	set_node_manufacturer_name = 7,
	set_node_product_name  = 8,
	set_node_name = 9,
	set_node_location = 10,

	set_value_label = 11,
	set_value_units = 12,
	set_value_help = 13,
	set_value = 14,
	refresh_value = 15,
	set_change_verified = 16,

    reset_controller = 17,
    soft_reset = 18,
    cancel_controller_command = 19,

    test_network_node  = 20,
    test_network  = 21,
    heal_network_node  = 22,
    heal_network  = 23,
    request_node_neighbor_update  = 24,
    assign_return_route  = 25,
    delete_all_return_routes  = 26,
    write_config  = 27,
    --polling z-wave devices
    set_poll_interval = 28,
    enable_poll = 29,
    disable_poll = 30,
    set_poll_intensity = 31,
mapping=50
}

function zwave_command(cmd)
	cmd = json.pdecode(cmd)
	if cmd then
		zwave.execcommand(cmd)
	end
end

function listen_commands()
	function ps()
		local rclient2 = redis.connect('127.0.0.1', 6379)
		for msg, abort in rclient2:pubsub({subscribe = {'zwave:commands'}}) do
		    if msg.kind == 'message' then
		     	zwave_command(msg.payload)
		    end
		end
	end
	coroutine.resume(coroutine.create(ps))
end

function get_commands()
	local cmd
	local row = db:getrow("SELECT * FROM zwave_commands ORDER BY id ASC LIMIT 1")
	if row then
		db:delete("zwave_commands", {id=row.id})
		cmd = json.pdecode(row.cmd)
		if cmd.cmd==50 then
			--refresh_mapping()
			return get_commands()
		end
	end
	return cmd
end

function _get_address(home_id, node_id, id, instance)
	local row = _get_row(_rname_value(home_id, node_id, id), instance)
	local a, w, address
	if row and row.address then
		address = row.address
		if address>0 then
			a = knxlib.decodega(address)
			w = row.is_writebus == 1
		end
	end
	return a, w, address
end
--вспомогательные для работы с hash таблицами с json с полями
function _get_row(name, key)
	local d = rclient:hget(name, key)
	if d then d = json.pdecode(d) end
	return d
end
function _set_row(name, key, fields)
	return rclient:hset(name, key, json.encode(fields))
end
function _remove_row(name, key)
	return rclient:hdel(name, key)
end
function _update_row(name, key, fields, create)
	local d = _get_row(name, key)
	if d then
		for k, v in pairs(fields) do d[k]=v end
	elseif create then
		d = fields
	end
	if d then _set_row(name, key, d) end
	return d
end


function driver_ready(home_id, version, node_id, cmode, suc_node_id, lib_type_name, lib_version, path, poll_interval)
	_update_row(_rname_driver, home_id, {
			version=version,
			node_id=node_id,
			cmode=cmode,
			suc_node_id=suc_node_id,
			lib_type_name=lib_type_name,
			lib_version=lib_version,
			path=path,
			poll_interval=poll_interval,
			state = 'Ready',
			refreshed = _get_time(),
			start_time = _get_time()
		}, true)
	hid = home_id
	_notifications({a="driver"})
	driver_present(home_id, nil, 60)
end

function driver_failed(home_id)
	_update_row(_rname_driver, home_id, {state='Failed', refreshed = _get_time()})
	_notifications({a="driver"})
end

function driver_present(home_id, c, time)
	if not time then time=6 end
	local n = _rname_present(home_id)
	rclient:set(n, 1)
	rclient:expire(n, time)
	--_update_row(_rname_driver, home_id, {refreshed=_get_time(), count=c})
end

function status_driver(home_id, cmode, state)
	_update_row(_rname_driver, home_id, {cmode=cmode, state=state, refreshed=_get_time()})
	_notifications({a="driver"})
end

function save_node(home_id, node_id,
is_listening, is_frequent_listening, is_beaming, is_routing, is_security,
max_baud_rate, version, is_zwaveplus,
node_type, manufacturer_name, product_name,
name, location, is_awake, is_failed
)
	_update_row(_rname_node(home_id), node_id, {
			is_listening = is_listening and 1 or 0,
			is_frequent_listening = is_frequent_listening and 1 or 0,
			is_beaming = is_beaming and 1 or 0,
			is_routing = is_routing and 1 or 0,
			is_security = is_security and 1 or 0,
			max_baud_rate = max_baud_rate,
			version = version,
				is_zwaveplus = is_zwaveplus and 1 or 0,
			node_type = node_type,
			manufacturer_name = manufacturer_name,
			product_name = product_name,
			name = name,
			location = location,
			is_awake = is_awake and 1 or 0,
			is_failed = is_failed and 1 or 0,
			refreshed = _get_time()
		}, true)
	_notifications({a="node", v=node_id})
end

function remove_node(home_id, node_id)
	--удаляем все связанные VALUE
	for _, id in ipairs(rclient:smembers(_rname_node_values(home_id, node_id))) do
		for _, instance in ipairs(rclient:hkeys(_rname_value(home_id, node_id, id))) do
			remove_value(home_id, node_id, nil, id, instance)
		end
	end
	_remove_row(_rname_node(home_id), node_id)
	_notifications({a="remove_node", v=node_id})
end

function name_node(home_id, node_id, name, manufacturer_name, product_name, location)
	_update_row(_rname_node(home_id), node_id, {
		manufacturer_name = manufacturer_name,
		product_name = product_name,
		location = location,
		name = name,
		refreshed = _get_time()
	})
	_notifications({a="node", v=node_id})
end


function status_node(home_id, node_id, is_awake, is_failed, is_timeout)
	_update_row(_rname_node(home_id), node_id, {
		is_awake = is_awake and 1 or 0,
		is_failed = is_failed and 1 or 0,
		is_timeout = is_timeout and 1 or 0,
		refreshed = _get_time()
	})

	_notifications({a="node", v=node_id})
end

--сохраняем конфигурация для значения
function save_value_configuraion(home_id, node_id, ind, id, instance,
genre, class, datatype,
label, units, help,
min, max,
is_readonly, is_writeonly, is_valueset, is_value_polled, is_polled, polled_intensity,
items, precision)
	local value = {
		ind = ind,
		genre = genre,
		class = class,
		datatype = datatype,
		label = label,
		units = units,
		help = help,
		min = min,
		max = max,
		is_readonly = is_readonly and 1 or 0,
		is_writeonly = is_writeonly and 1 or 0,
		is_valueset = is_valueset and 1 or 0,
		is_value_polled = is_value_polled and 1 or 0,
		is_polled = is_polled and 1 or 0,
		polled_intensity = polled_intensity,
		items = json.encode(items),
		precision = precision,
		refreshed = _get_time()
	}
	_update_row(_rname_value(home_id, node_id, id), instance, value, true)
	rclient:sadd(_rname_node_values(home_id, node_id), id)
	_notifications({a="value", id=id, node_id=node_id, instance=instance})
end

function remove_value(home_id, node_id, index, id, instance)
	--удаляем из маппинга
	local address
	_, _, address = _get_address(home_id, node_id, id, instance)
	if address then
		_remove_row(_rname_address(address), _rname_address_hkey(home_id, node_id, id, instance))
	end
	--удалаем саму запись
	_remove_row(_rname_value(home_id, node_id, id), instance)
	--удаляем связку с node
	rclient:srem(_rname_node_values(home_id, node_id), id)
end

--включен/выключен опрос
function polling_value(home_id, node_id, ind, id, instance, is_polled)
	_update_row(
		_rname_value(home_id, node_id, id),
		instance,
		{is_polled=is_polled and 1 or 0, refreshed = _get_time()}
	)
	_notifications({a="value", id=id, node_id=node_id, instance=instance})
end

--сохраняем конкретное значение в групповой адрес
function set_value(home_id, node_id, ind, id, instance, datatype, value)
	_update_row(
		_rname_value(home_id, node_id, id),
		instance,
		{value=json.encode(value), refreshed = _get_time()}
	)
	local address, writebus = _get_address(home_id, node_id, id, instance)
	if address then
		_notifications({a="setvalue", id=id, node_id=node_id, instance=instance, dt=datatype, value=json.encode(value)})
		if (type(value)=="table") then --list
			local items = value
			value = 0
			for k, v in pairs(items) do
				value = k
				break
			end
		end

		local vvv = grp.getvalue(address)
		if vvv~=value then
			if writebus then
			    grp.write(address, value)
			else
			    grp.update(address, value)
			end
		end
	end
end


function init()
	--создаем таблицы, если не созданы
	create_table_zwave()
	create_table_nodes()
	create_table_values()
	create_table_commands()
	rclient:del(_rname_notifications)
--	db:query("DELETE FROM zwave_commands")
end

function close()
	if hid then
		_update_row(_rname_driver, hid, {state='Close', refreshed=_get_time()})
	end
end

local datatypes = {}
-- bus write/response
function eventhandler(event)
	if type(event)=="table" then
		local id = event.dstraw

		local name = _rname_address(id)
		if rclient:exists(name) then
			for _, v in ipairs(rclient:hvals(name)) do
				v = json.pdecode(v)
				if v  and (not v.is_readonly or v.is_readonly==0) then
					local a = event.dst
					local _dt = datatypes[a]

					if _dt==nil then
						local o = grp.find(a)
						if o then
							_dt = o.datatype
							datatypes[a] = _dt
						end
					end

					if dt~=nil then
						zwave.writevalue(v.node_id, v.id, v.instance,  knxdatatype.decode(event.datahex, _dt))
					end
				end
			end
		end
	end
end

client = require('localbus').new(1)
client:sethandler('groupwrite', eventhandler)

function getlbfd()
	return client.sock:getfd()
end

function step()
	if client then
		client:step()
	end
end
